/* APACHETOP
**
**
*/
#include "apachetop.h"

struct gstat gstats;
time_t now;

Circle *c;
Timed_Circle *tc;
Hits_Circle *hc;

map *um, /* urlmap */
    *im, /* ipmap */
    *hm, /* hostmap */
    *rm; /* referrermap */

struct config cf;

int dprintf(const char *fmt, ...)
{
	FILE *d;
        va_list args;

	if (cf.debug && (d = fopen(DEBUG_OUTPUT, "a")))
	{
        	va_start(args, fmt);
        	vfprintf(d, fmt, args);
		fclose(d);
        	va_end(args);
	}

        return 0;
}

int recordstats(struct logbits l)
{
	int t;

	/* global stats */
	if (gstats.alltime.first == 0) gstats.alltime.first = now;
	gstats.alltime.last = now;
	gstats.alltime.reqcount++;
	gstats.alltime.bytecount += l.bytes;

	/* for this return code, increment */
	t = (int)(l.retcode/100);
	gstats.r_codes[t].reqcount++;
	gstats.r_codes[t].bytecount += l.bytes;

	/* url and ip specific stuff was done in the parser */

	return 0;
}

float readableNum(long long num, char *suffix)
{
	#define MAX_IN_KB (1024*1024)*10

	if (num < MAX_IN_KB)
	{
		*suffix = 'K';
		return (float)num/1024;
	}

	*suffix = 'M';
	return (float)num/(1024*1024);
}

void display()
{
	unsigned int diff, d = 0, h = 0, m = 0, s = 0;
	float bytes, ftmp;
	int x, disp;
	long ltmp;
	char ctmp[80], bytes_suffix, *cptmp;
	struct tm *tm;
	
	/* for easy reference during walk() */
	map *map; unsigned int pos;

	/* walk() pointer */
	struct logbits *lb;

	/* last hit and uptime */
	tm = localtime(&gstats.alltime.last);
	strftime(ctmp, sizeof(ctmp), "%H:%M:%S", tm);
	diff = (unsigned int)difftime(now, gstats.start);
	if (diff > 86399) diff -= ((d = diff / 86400)*86400);
	if (diff > 3599) diff -= ((h = diff / 3600)*3600);
	if (diff > 59) diff -= ((m = diff / 60)*60);
	s = diff;
	mvprintw(0, 0,
	  "last hit: %s        atop runtime: %2d days, %02d:%02d:%02d",
	  ctmp, d, h, m, s );

	/* current time */
	tm = localtime(&now);
	strftime(ctmp, sizeof(ctmp), "%H:%M:%S", tm);
	mvprintw(0, 71, "%s", ctmp);

	//All: 1,140,532 requests (39.45/sec),  999,540,593 bytes (857,235/sec) 
	bytes = readableNum(gstats.alltime.bytecount, &bytes_suffix);
	ftmp = getMAX(now-gstats.alltime.first, 1); /* divide-by-zero hack */
	attron(A_BOLD);
	mvprintw(1, 0, 
	  "All: %12lld reqs (%6.1f/sec)  %11.1f%c (%6.1fK/sec)  (%7.1fK/req)",
	  gstats.alltime.reqcount,
	  gstats.alltime.reqcount/ftmp,
	  bytes, bytes_suffix,
	  (gstats.alltime.bytecount/1024)/ftmp,
	  (float)(gstats.alltime.bytecount/1024)/
	      getMAX(gstats.alltime.reqcount, 1));
	attroff(A_BOLD);


	// 2xx 1,604,104 (95%) 3xx 1,000,000 ( 3%) 4xx 1,000,000 ( 1%) 
	// 5xx 1,000,000 ( 1%)
	ftmp = gstats.r_codes[2].reqcount + gstats.r_codes[3].reqcount+
	       gstats.r_codes[4].reqcount + gstats.r_codes[5].reqcount;
	if (ftmp == 0) ftmp = 1; /* avoid NaN with no hits */
	mvprintw(2, 0,
	  "2xx: %7lld (%4.1f%%) 3xx: %7lld (%4.1f%%) "
	  "4xx: %5lld (%4.1f%%) 5xx: %5lld (%4.1f%%) ",
	  gstats.r_codes[2].reqcount, (gstats.r_codes[2].reqcount/ftmp)*100,
	  gstats.r_codes[3].reqcount, (gstats.r_codes[3].reqcount/ftmp)*100,
	  gstats.r_codes[4].reqcount, (gstats.r_codes[4].reqcount/ftmp)*100,
	  gstats.r_codes[5].reqcount, (gstats.r_codes[5].reqcount/ftmp)*100);

	/* housecleaning on the circle, if its required in this class */
	c->updatestats();
	/* fetch the time of the first "recent" request */
	ltmp = now - c->oldest();
	ltmp = getMAX(ltmp, 1); /* divide-by-zero hack */
	bytes = readableNum(c->getbytecount(), &bytes_suffix);
	attron(A_BOLD);
	mvprintw(3, 0,
	  "R (%3ds): %7lld reqs (%6.1f/sec)  %11.1f%c (%6.1fK/sec)  (%7.1fK/req)",
	  ltmp, c->getreqcount(),
	  ((float)c->getreqcount()/ltmp),
	  bytes, bytes_suffix,
	  ((float)(c->getbytecount()/1024)/ltmp),
	  (float)(c->getbytecount()/1024)/getMAX(c->getreqcount(), 1)
	  );
	attroff(A_BOLD);

	ftmp = c->getsummary(2) + c->getsummary(3) +
	       c->getsummary(4) + c->getsummary(5);
	if (ftmp == 0) ftmp = 1; /* avoid NaN with no hits */
	mvprintw(4, 0,
	  "2xx: %7lld (%4.1f%%) 3xx: %7lld (%4.1f%%) "
	  "4xx: %5lld (%4.1f%%) 5xx: %5lld (%4.1f%%) ",
	  c->getsummary(2), (c->getsummary(2)/ftmp)*100,
	  c->getsummary(3), (c->getsummary(3)/ftmp)*100,
	  c->getsummary(4), (c->getsummary(4)/ftmp)*100,
	  c->getsummary(5), (c->getsummary(5)/ftmp)*100);

	/* top $X urls or ips/hosts */

	/* make an array containing all we have, then sort it */

	OAHash item_hash;
	int item_used = 0, items_size;
	struct peritem_stat *items = NULL, *item_ptr = NULL;

	items_size = c->getreqcount();

	item_hash.create(items_size*5);

	/* this is our array */
	items = (struct peritem_stat *)
	    calloc(items_size * sizeof(peritem_stat), 1);
	
	/* pick a map to use */
	switch(cf.display_mode)
	{
		default:
		case DISPLAY_URLS: map = um; break;
		case DISPLAY_IPS: map = im; break;
		case DISPLAY_REFS: map = rm; break;
	}

	while(c->walk(&lb) != -1)
	{
		/* skip unused */
		if (lb == NULL)
			continue;
	
		/* FILTERS */

		/* set up some pointers, depending on cf.display_mode, so
		 * that we can refer to the url_pos or ip_pos via just pos,
		 * and the urlmap or ipmap via just map
		*/
		switch(cf.display_mode)
		{
			default:
			case DISPLAY_URLS: pos = lb->url_pos; break;
			case DISPLAY_IPS: pos = lb->ip_pos; break;
			case DISPLAY_REFS: pos = lb->ref_pos; break;
		}

		cptmp = map->reverse(pos);
		item_ptr = (struct peritem_stat *)item_hash.lookup(cptmp);
		
		if (item_ptr == NULL)
		{
			/* not seen it, make a new slot */
			item_ptr = &items[item_used];

			item_hash.insert(map->reverse(pos), &items[item_used]);
			item_ptr->item = pos;

			++item_used;
			item_ptr->last = now;
			item_ptr->first = lb->time;
		}

		/* update stats for this item in the array */
		++(item_ptr->reqcount);
		item_ptr->bytecount += lb->bytes;

		if (lb->time < item_ptr->first)
			item_ptr->first = lb->time;
	}

	/* we're done with this */
	item_hash.destroy();

	/* calculate some stats; timespan, kbps, rps */
	for(x = 0 ; x < item_used ; ++x)
	{
		item_ptr = &items[x];

		item_ptr->timespan = item_ptr->last - item_ptr->first;
		if (item_ptr->timespan == 0)
			item_ptr->timespan = 1;

		item_ptr->kbps = ((float)
		    (item_ptr->bytecount/1024)/item_ptr->timespan);
		item_ptr->rps = ((float)
		    item_ptr->reqcount/item_ptr->timespan);
	}

#if 0
	/* what are we showing in the table? we have a few options;
	 *
	 * top $X URLS or IPs/Hosts or Referrers
	 * detailed referrer stats for a given URL
	*/
	if (cf.display_mode == DISPLAY_URLS && cf.selected_item_pos)
	{
		/* detailed referrer stats for a given URL */
		clrtobot();
		mvprintw(LINES_RESERVED-1, 0, "REQS REQ/S    KB  KB/S URL");

		/* find the url_pos given by display_referrer in items */
		for (x = 0 ; x < item_used ; ++x)
		{
			item_ptr = &items[x];
			if (item_ptr->item == cf.selected_item_pos)
			{
				/* show stats for this line only */
				showDetail(item_ptr, 0);
				break;
			}
		}

		/* now display breakdown of referrers for that line */
		//displayReferrers();
		return;
	}
#endif


	/* top $X items; sort the array for display */

	/* default is reqcount_cmp (avoids compiler warning) */
	int (*whichsort)(const void *, const void *) = reqcount_cmp;
	/* what are we sorting by? */
	switch(cf.sort)
	{
		case SORT_REQCOUNT: whichsort = reqcount_cmp; break;
		case SORT_REQPERSEC: whichsort = rps_cmp; break;
		case SORT_BYTECOUNT: whichsort = bytes_cmp; break;
		case SORT_BYTESPERSEC: whichsort = bps_cmp; break;
	}
	qsort(items, item_used, sizeof(struct peritem_stat), whichsort);

	clrtobot();

	/* display something */

	/* a suitable header, at LINES_RESERVED-1
	 * (-1 because curses starts at zero, but I've started at 1) */
	switch(cf.display_mode)
	{
		case DISPLAY_URLS:
		mvprintw(LINES_RESERVED-1, 0, "REQS REQ/S    KB  KB/S URL");
		break;
	
		case DISPLAY_IPS:
		mvprintw(LINES_RESERVED-1, 0, "REQS REQ/S    KB  KB/S HOST");
		break;

		case DISPLAY_REFS:
		mvprintw(LINES_RESERVED-1, 0, "REQS REQ/S    KB  KB/S REFERRER");
		break;
	}

	disp = 0; /* count how many we've shown */
	for(x = 0, item_ptr = &items[0] ;
	    x < items_size  &&  disp < LINES-LINES_RESERVED ;
	    ++item_ptr, ++x)
	{
		/* translate a table position (0 to ..) into an url_pos,
		 * ip_pos, or ref_pos, depending on current display mode. We
		 * can then use selected_item_pos to get info for the
		 * selected item if the user hits return to get it */
		if (cf.selected_item == x)
			cf.selected_item_pos = item_ptr->item;
	
		/* skip empty tablespaces, even though none should exist */
		if (item_ptr->reqcount == 0)
			continue;

		/* render the line itself at position disp. map should be
		 * already set from earlier in this function */
		showDetail(item_ptr, disp, map);

		++disp;
	}
	free(items);

	cf.current_display_size = disp;

	/* if there were items, draw a marker */
	if (cf.current_display_size)
		drawMarker();
	else
	{
		/* come to rest just above the column headers */
		move(LINES_RESERVED-2,0);

		refresh();
	}
}

/* update position of asterisk next to URLs */
void drawMarker(void)
{
	if (cf.current_display_size == 0)
		return;

	/* ensure our marker isn't beyond the end of the list */
	if (cf.selected_item > cf.current_display_size-1)
		cf.selected_item = cf.current_display_size-1;
	
	/* or above the start of it */
	if (cf.selected_item < 0) cf.selected_item = 0;

	/* draw an asterisk next to the selected item and clear adjacent lines*/
	mvaddch(LINES_RESERVED+cf.selected_item-1, COLS_RESERVED-3, ' ');
	mvaddch(LINES_RESERVED+cf.selected_item, COLS_RESERVED-3, '*' | A_BOLD);
	mvaddch(LINES_RESERVED+cf.selected_item+1, COLS_RESERVED-3, ' ');

	/* come to rest just above the column headers */
	move(LINES_RESERVED-2,0);
	
	refresh();
}

#define REALCMP(a, b) (a == b) ? 0 : (a < b  ? 1 : -1)
/* perurl comparison functions {{{ */
int reqcount_cmp(const void *a, const void *b)
{
	return REALCMP(((struct peritem_stat *)a)->reqcount,
	               ((struct peritem_stat *)b)->reqcount);
}
int rps_cmp(const void *a, const void *b)
{
	return REALCMP(((struct peritem_stat *)a)->rps,
	               ((struct peritem_stat *)b)->rps);
}
int bytes_cmp(const void *a, const void *b)
{
	return REALCMP(((struct peritem_stat *)a)->bytecount,
	               ((struct peritem_stat *)b)->bytecount);
}
int bps_cmp(const void *a, const void *b)
{
	return REALCMP(((struct peritem_stat *)a)->kbps,
	               ((struct peritem_stat *)b)->kbps);
} /* }}} */

/* show all referrers for the given item_ptr's urlpos */
void displayReferrers(struct peritem_stat *item_ptr)
{
}

/* render contents of item_ptr into a statistics line onscreen,
 * at offset vert_location; pull the text portion out of m */
void showDetail(struct peritem_stat *item_ptr, int vert_location, map *m)
{
	/* start drawing at LINES_RESERVED */
	mvprintw(LINES_RESERVED+vert_location, 0,
	    "%4lld %5.*f %5.*f %5.*f %.*s",
	    item_ptr->reqcount,

	    ((item_ptr->rps > 99) ? 1 : 2),
	    item_ptr->rps,

	    /* scale KB display; if >1000K lose the decimal point */
	    ((item_ptr->bytecount > (999*1024)) ? 0 : 1),
	    ((float)item_ptr->bytecount/1024),
	    
	    /* scale KB/s display; if >100K/s lose the decimal point */
	    ((item_ptr->kbps > 99) ? 0 : 1),
	    item_ptr->kbps,

	    (COLS-COLS_RESERVED /* width */), m->reverse(item_ptr->item));
}

static void finish(int s)
{
	endwin();

	exit(0);
}

int main(int argc, char *argv[])
{
	int fd_in, buflen, ch, x;
	char buf[32768], *bufcp, *nl;
	time_t last_display = 0;
	struct logbits lb;
	struct timeval stv;
	bool seen_h = 0, seen_t = 0;
	WINDOW *win;
	off_t offset = 0, prev_offset = 0;
	extern char *optarg;

//	pthread_t resolve_thread;

	LogParser *p;
	CommonLogParser *cmmp; //AtopLogParser *atpp;

	stv.tv_sec = 0; stv.tv_usec = 200000;

	/* some init stuff will need to know the time */
	now = time(NULL);

	/* set up initial configuration {{{ */
	cf.debug = true;
	cf.current_display_size = 0;
	cf.logfile = strdup(DEFAULT_LOGFILE);
	cf.logtype = DEFAULT_LOGTYPE;
	cf.circle_size = DEFAULT_CIRCLE_SIZE;
	cf.circle_mode = DEFAULT_CIRCLE_MODE;
	cf.sort = DEFAULT_SORT;
	cf.refresh_delay = DEFAULT_REFRESH_DELAY;
	cf.display_mode = DEFAULT_DISPLAY_MODE;
	cf.display_paused = false;
	cf.keep_querystring = false;
	cf.lowercase_urls = false;
	cf.keep_segments = 0;
	cf.preserve_ref_protocol = 0; /* }}} */

	/* process commandline */
	while ((ch = getopt(argc, argv, "f:t:H:T:hqls:pr:")) != -1)
	{
		switch(ch)
		{
			case 'f':
				free(cf.logfile);
				cf.logfile = strdup(optarg);
				break;
			case 't':
				if (strcmp(optarg, "common") == 0)
					cf.logtype = LOG_COMMON;
				else if (strcmp(optarg, "combined") == 0)
					cf.logtype = LOG_COMBINED;
				else if (strcmp(optarg, "atop") == 0)
					cf.logtype = LOG_ATOP;
				break;
			case 'T':
				x = atoi(optarg);
				if (x > 0)
				{
					cf.circle_mode = TIMED_CIRCLE;
					cf.circle_size = x;
					seen_t = 1;
				}
				break;
			case 'H':
				x = atoi(optarg);
				if (x > 0)
				{
					cf.circle_mode = HITS_CIRCLE;
					cf.circle_size = x;
					seen_h = 1;
				}
				break;

			case 'q':
				cf.keep_querystring = true;
				break;

			case 'l':
				cf.lowercase_urls = true;
				break;

			case 's':
				x = atoi(optarg);
				if (x > 0)
					cf.keep_segments = x;
				break;

			case 'p':
				cf.preserve_ref_protocol = 1;
				break;

			case 'r':
				x = atoi(optarg);
				if (x > 0)
					cf.refresh_delay = x;
				break;

			case 'h': 
			case '?': 
				usage(argv[0]);
				exit(1);
				break;
		}
	}

	if (seen_t && seen_h)
	{
		fprintf(stderr, "%s: -t and -h are mutually exclusive."
		    "Specify only one.\n", argv[0]);
		exit(1);
	}
	
	
	if ((fd_in = open(cf.logfile, O_RDONLY)) == -1)
	{
		usage(argv[0]);
		fprintf(stderr, "Opening logfile %s:\n%s\n",
		    cf.logfile, strerror(errno));
		exit(1);
	}

	/* remember cf.circle_size hits */
	switch(cf.circle_mode)
	{
		default:
		case HITS_CIRCLE:
			hc = new Hits_Circle;
			hc->create(cf.circle_size);
			c = hc;
			break;
	
		case TIMED_CIRCLE:
			tc = new Timed_Circle;
			tc->create(cf.circle_size);
			c = tc;
			break;
	}

	switch(cf.logtype)
	{
		default:
		case LOG_COMMON:
		case LOG_COMBINED: /* CommonLogParser handles combined too */
			cmmp = new CommonLogParser;
			p = cmmp;
			break;
#if 0
/* this isn't written yet */
		case LOG_ATOP:
			atpp = new AtopLogParser;
			p = atpp;
			break;
#endif
	}

	/* maps auto-resize, just use a sane starting point */
	/* url string -> url hash map */
	um = new map; /* hashing happens internally */
	um->create(cf.circle_size);

	/* referrer string -> hash map */
	rm = new map;
	rm->create(cf.circle_size);

	/* ip string -> ip hash map */
	im = new map; 
	im->create(cf.circle_size);

	/* host string -> host hash map */
	hm = new map; 
	hm->create(cf.circle_size);

	//queue = new Queue;

	memset(&gstats, (char) NULL, sizeof(gstats));
	gstats.start = time(NULL);

	signal(SIGINT, finish);
	signal(SIGKILL, finish);

	initscr();
	win = newwin(0,0,0,0);
	nonl(); /* no NL->CR/NL on output */
	cbreak(); /* enable reading chars one at a time */
	noecho();
	keypad(win, true);

	nodelay(win, true);

	/* set off a thread to do DNS resolving in the background */
	//pthread_create(&resolve_thread, pthread_attr_default, &resolver, NULL);

	for(;;)
	{
		now = time(NULL);

		if ((ch = wgetch(win)) != ERR)
			readKey(ch);

		/* only update display every 2 seconds */
		if (!cf.display_paused &&
		    now - last_display >= cf.refresh_delay)
		{
			display();
			last_display = now; // remember our last update
		}

		offset = lseek(fd_in, 0, SEEK_END);

		if (prev_offset == 0)
			/* this is our first seek, ignore it */
			prev_offset = offset;

		if (offset <= prev_offset)
		{
			/* nothing new to read. if we didn't receive a
			 * keypress (ch == ERR) then delay for a while. The
			 * ch check is to avoid the feeling of slowdown if
			 * the user is repeatedly keying-up or refresh etc */
			if (ch == ERR) select(0, NULL, NULL, NULL, &stv);
			prev_offset = offset;
			continue;
		}

		/* seek back to where we last read to */
		lseek(fd_in, prev_offset, SEEK_SET);

		/* read some data */
		buflen = read(fd_in, buf, sizeof(buf));

		prev_offset = offset;

		/* pass each line to logsplit */
		nl = bufcp = buf;
		while ((nl - buf) < buflen && nl)
		{
			bufcp = nl;
			/* find the end of this line */
			if ((nl = strchr(bufcp, '\n')))
			{
				*nl = (char) NULL;
				++nl;

				if (p->parse(bufcp, &lb) == 0)
				{
					/* insert into circle */
					c->insert(lb);

					/* record stats */
					recordstats(lb);
				}
			}

			// next line..
		}
	}

	/* NOTREACHED */
	return 0;
}
#if 0
/* resolver stuff, unused {{{ */
/* background thread that watches ipmap and resolves IPs in it to hosts if
 * they don't have an entriy in the hostmap; and vice versa.
 *
 * entries between the maps are linked via their pos; they have the same one.
*/
void *resolver(void *)
{
	int rc;
	struct qresult *r;
	struct timeval stv;

	/* some tmp buffers for resolve() to fill in */
	char *host, ip[256];
	int ipl;

	stv.tv_sec = 2;
	stv.tv_usec = 0;

	/* run sequence:
	**
	** every few seconds:
	** loop through ipmap (write walk()) getting the pos for each one
	** reverse that pos in hostmap and see if it's NULL
	**  if it is, do gethost crap to resolve it
	**  store result
	*/
	do
	{
		/* wait for entries to appear in the queue, then process */
		if (queue->entries() && (rc = queue->pop( (void **)&r )) == 0)
		{
			dprintf("got %s to check\n", r->src);

			resolve(r->src, &ipl, (char **)&ip, &host);

			dprintf("host of that is %s\n", host);
			/* do what now? */

			free(r->src);
			free(r);
		}
		else
		{
			dprintf("sleep\n");
			/* zzz */
			select(0, NULL, NULL, NULL, &stv);
		}
	}
	while (queue->entries() > 0 || main_done == 0);

	return 0;
}
/* given an unknown source, populate the ipl, ip, and host fields, where
 * possible; 
 * host must be a pointer, ip must be a char[].
*/
int resolve(char *src, int *ipl, char **ip, char **host)
{
	in_addr_t f;
	struct hostent *he;

	if ((f = inet_addr(src)) == INADDR_NONE)
	{
		/* its a host */
		if ((he = gethostbyname(src)))
		{
			*host = src;
			inet_ntop(AF_INET, he->h_addr, *ip, sizeof(*ip));
			*ipl = inet_addr(*ip);
		}
		else
		{
			/* it didn't resolve; this shouldn't happen because
			 * Apache shouldn't put duff DNS in its logs */
			*host = src;
			*ipl = 0;
			*ip = NULL;
		}
	}
	else
	{
		/* its an IP */
		*ipl = f;
		if ((he = gethostbyaddr((char *)&f, sizeof(f), AF_INET)))
		{
			*host = he->h_name;
			inet_ntop(AF_INET, he->h_addr, *ip, sizeof(*ip));
		}
		else
		{
			/* no reverse, this is fine */
			*host = NULL;
			*ip = src;
		}
	}
	//printf("host: %s ip: %s ipl: %d\n", *host, *ip, *ipl);

	return 0;
}
/* }}} */
#endif

int readKey(int ch)
{
	/* got a keypress, best process it */
	switch(ch)
	{
		case 'p': /* (un)pause display */
			cf.display_paused = !cf.display_paused;
			if (cf.display_paused)
			{
				/* tell the user we're paused */
				attron(A_REVERSE);
				mvprintw(0, 60, "paused");
				attroff(A_REVERSE);
				refresh();
			}
			else
			{
				/* turning off pause forces an update */
				move(0, 60); /* clear "paused" */
				clrtoeol(); /* time will get redrawn */
				display();
			}
			break;
	
		case 'r': /* sorting */
			cf.sort = SORT_REQCOUNT;
			break;
		case 'R':
			cf.sort = SORT_REQPERSEC;
			break;
		case 'b':
			cf.sort = SORT_BYTECOUNT;
			break;
		case 'B':
			cf.sort = SORT_BYTESPERSEC;
			break;
#if 0
		case 's': /* new refresh delay */ 
#ifndef SOLARIS /* linking against readline is failing on Solaris atm */
			endwin();
			char *t;
			int nd;
			t = readline("Seconds to Delay: ");
			if ((nd = atoi(t)) != 0)
				cf.refresh_delay = nd;

			refresh();
#endif
			break;
#endif
		case 'd': /* display mode; urls or hosts */
			switch(cf.display_mode)
			{
			case DISPLAY_URLS:
				cf.display_mode = DISPLAY_IPS;
				break;
			case DISPLAY_IPS:
				cf.display_mode = DISPLAY_REFS;
				break;
			case DISPLAY_REFS:
				cf.display_mode = DISPLAY_URLS;
				break;
			}
			break;

		case KEY_UP:
			/* sanity checking for this is done in drawMarker() */
			cf.selected_item--;
			drawMarker();
			break;

		case KEY_DOWN:
			/* sanity checking for this is done in drawMarker() */
			cf.selected_item++;
			drawMarker();
			break;
		
		case KEY_REFRESH:
			refresh();
			break;

		case KEY_BREAK:
		case 'q':
			finish(0);
	}
	
	return 0;
}

void usage(char *cmd)
{
	fprintf(stderr, 
	    "Usage: %s [-qlp] [-f file] [-t logtype] [-s num] [-H hits | -T time]\n"
	    "File options:\n"
	    "  -f logfile  open logfile [%s]\n"
	    "  -t logtype  specify logfile type [common]\n"
	    "\n"
	    "URL/referrer munging options:\n"
	    "  -q          keep query strings [no]\n"
	    "  -l          lowercase all URLs [no]\n"
	    "  -s num      keep num path segments of URL [all]\n"
	    "  -p          preserve protocol at front of referrers [no]\n"
	    "\n"
	    "Stats options:\n"
	    "  Supply up to one of the following two. default: [-%c %d]\n"
	    "  -H hits     remember stats for this many hits\n"
	    "  -T secs     remember stats for this many seconds\n"
	    "\n"
	    "  -r secs     refresh delay in seconds [10]\n"
	    "\n"
	    "  -h          this help\n"
	    "\n"
	    ,
	    cmd,
	    DEFAULT_LOGFILE,
	    DEFAULT_CIRCLE_MODE, DEFAULT_CIRCLE_SIZE
	    );

	return;
}
